/* vim: se fo=tcroq : */
/* Copyright (C) 2016 Penguin Development
 *
 * This file is part of Ising, a D-dimensional Ising model simulator.
 *
 * License GPLv3+: GNU GPL version 3 or later
 * 	<http://gnu.org/licenses/gpl.html>.
 * This is free software: you are free to change and redistribute it.
 * There is NO WARRANTY, to the extent permitted by law.
 *
 * Written by Peter Sterk.
 */
#ifndef ISING_H
#define ISING_H

#include <atomic>
#include <algorithm>
#include <array>
#include <chrono>
#include <cmath>
#include <iostream>
#include <functional>
#include <random>
#include <sstream>
#include <stack>
#include <stdexcept>

#ifdef HAVE_BOOST_FILESYSTEM_PATH_HPP
#include <boost/filesystem.hpp>
#include <boost/filesystem/fstream.hpp>
namespace fs = boost::filesystem;
#endif

#include <ising/utility.h>

template<long long int D, long long int L> class Ising
{
    public:
        /** Structure representing a macroscopic state. This is a simple
         * data structure holding all the static and dynamic system variables:
         * inter-spin interaction strength, field strength, temperature,
         * magnetisation and exchange energy.
         */
        struct MacroState
        {
            public:
                MacroState (void) : J (0), h (0), T (0), M (0), EJ (0) {}
                ~MacroState (void) {}
                double J, h, T;
                long long int M, EJ;

                /** Utility method to calculate the energy of the macrostate.
                 * @return -J*EJ-h*M
                 */
                inline double E (void) const { return -J*EJ-h*M; }
                friend inline std::ostream &operator<< (std::ostream &os,
                    const MacroState &st)
                {
                    os << st.J << " " << st.h << " " << st.T << " " << st.M <<
                        " " << st.EJ << " " << st.E ();
                    return os;
                }
        };

        /** Constructor with initial conditions J, h and T and an initial
         * lattice generator function.
         * @param J Inter-spin interaction strength in Kelvin.
         * @param h field strength in Kelvin.
         * @param T temperature in Kelvin.
         * @param genfunc Lattice generation function. This should only ever
         * return +1 or -1, otherwise Bad Things™ will happen. Default:
         * inftemp_generator.
         */
        Ising<D, L> (double J = 1.0, double h = 0.0, double T = 289.15,
                std::function<int8_t(void)> genfunc = inftemp_generator);
        ~Ising<D, L> (void) {}

        /** Perform one Metropolis algorithm trial.
         * @return true if the trial was accepted, false otherwise.
         */
        bool trial (void);

        /** Run the simulation for multiple Metropolis trials.
         * @param iterations number of trials to run the simulation for.
         * @param trial_limit threshold of successive failed trials before a
         * steady-state (ground state) is accepted.
         * @param macrostream pointer to stream to which to output macroscale
         * data every saveafter_macro trials.
         * @param saveafter_macro number of trials after which to output
         * macroscale data to macrostream.
         * @param saveafter number of iterations after which a full
         * state-snapshot is saved, if savedir is given. The first and final
         * states are always saved, provided savedir is set.
         * @param savedir directory into which state snapshots should be
         * saved. Set to a zero-length string if no snapshots should be saved
         * (macroscale data is always output to macrostream regardless of
         * this).
         * @param progstream pointer to stream to which to report progress in
         * 1% intervals.
         * @return the number of iterations run.
         */
        std::size_t run (
            std::size_t iterations = ~(static_cast<std::size_t> (0)) - 1,
            std::size_t trial_limit = 10000,
            std::ostream *macrostream = nullptr,
            std::size_t saveafter_macro = 1,
            std::size_t saveafter = 1000,
            const std::string &savedir = "",
            std::ostream *progstream = nullptr);

        /** Perform a single clustered flip using the Wolff algorithm.
         * @return the number of spins flipped.
         */
        std::size_t wolff (void);

        /** Run the simulation for multiple Wolff steps.
         * @param iterations number of iterations to run the simulation for.
         * @param macrostream pointer to stream to which to output macroscale
         * data every generation.
         * @param saveafter number of iterations after which a full
         * state-snapshot is saved, if savedir is given. The first and final
         * states are always saved, provided savedir is set.
         * @param savedir directory into which state snapshots should be
         * saved. Set to a zero-length string if no snapshots should be saved
         * (macroscale data is always output to macrostream regardless of
         * this).
         * @param progstream pointer to stream to which to report progress in
         * 1% intervals.
         * @return the number of iterations run.
         */
        std::size_t wolff_run (
            std::size_t iterations = ~(static_cast<std::size_t> (0)) - 1,
            std::ostream *macrostream = nullptr,
            std::size_t saveafter = 1000,
            const std::string &savedir = "",
            std::ostream *progstream = nullptr);

#ifdef HAVE_BOOST_FILESYSTEM_PATH_HPP
        /** Save the current microstate to the given file.
         * @param f Path to the file into which the microstate is to be
         * stored.
         */
        void save (const fs::path &f);
#endif

        /** Produce a stringified (`graphical') x-slice (1D) or (x, y)-slice
         * (2D+) at the given higher-dimensional coordinates.
         * @param coords z, w, ... coordinates of the slice. coords should
         * have length 0 in 1D and length D-2 in 2D and higher.
         * @param upchar Character for spin-up (default: '+')
         * @param downchar Character for spin-down (default: '-')
         * @return A string of the characters upchar or downchar for each spin
         * in the slice. The line number indicates the y coordinate, while the
         * character number in the line is the x coordinate.
         */
        std::string xyslice (
            const std::array<long long int, D == 1 ? 0 : (D - 2)> &coords,
            char upchar = '+', char downchar = '-');

        /** Set the inter-spin interaction strength in Kelvin (because
         * Boltzmann's constant is set to 1). Positive J favours alignment of
         * spins.
         * @param J the desired inter-spin interaction strength in Kelvin.
         */
        void set_J (double J);
        /** Set the field strength/per-spin energy in Kelvin (because
         * Boltzmann's constant is set to 1). Positive h favours spin-up.
         * @param h the desired field strength in Kelvin.
         */
        void set_h (double h);
        /** Set the system temperature in Kelvin.
         * @param T the desired system temperature in Kelvin.
         */
        void set_T (double T);

        /** Get the current macrostate.
         * @return the current macrostate.
         */
        MacroState get_macrostate (void) const noexcept { return state; }

        /** Check whether run() is currently active, e.g. from a different
         * thread.
         * @return true if run() is busy, false otherwise.
         */
        bool is_running (void) { return running; }

        /** Infinite-temperature initial lattice generator; returns +1 or -1
         * with equal probability.
         * @return +1 or -1, with 50% probability.
         */
        static int8_t inftemp_generator (void);
        /** Zero-temperature initial lattice generator; always returns 1.
         * @return 1;
         */
        static int8_t zerotemp_generator (void) { return 1; }

        /** Generator function class that populates the spin field from
         * a previously saved microstate, streamed from an std::istream.
         */
        class StreamLoaderGenerator
        {
            public:
                StreamLoaderGenerator (void) = delete;
                StreamLoaderGenerator (std::istream &stream) : is (stream) {}
                ~StreamLoaderGenerator (void) {}

                int8_t operator() (void)
                {
                    return (is.get () == '+') ? 1 : -1;
                }

            protected:
                std::istream &is;
        };

        /** Generator function class that populates the spin field from
         * a previously saved microstate stored in a string.
         */
        class StringLoaderGenerator
        {
            public:
                StringLoaderGenerator (void) = delete;
                StringLoaderGenerator (std::string &str) : s (str), i (0)
                {
                    if (s.length () < cstpow (L, D))
                        throw std::length_error (std::string (
                            "String must have a length of a least ") +
                            std::to_string (cstpow (L, D)));
                }
                ~StringLoaderGenerator (void) {}

                int8_t operator() (void)
                {
                    return (s[i++] == '+') ? 1 : -1;
                }

            protected:
                std::string &s;
                std::size_t i;
        };

        std::atomic<bool> interrupt;

        static constexpr long long int dim = D;
        static constexpr long long int size = L;
        static constexpr long long int N = cstpow (L, D);

    protected:
        std::array<int8_t, cstpow (L, D)> spins;
        MacroState state;
        std::atomic<bool> running;

        std::default_random_engine randgen;
        std::uniform_int_distribution<long long int> spinpicker;
        std::uniform_real_distribution<double> acceptor;

#ifdef HAVE_BOOST_FILESYSTEM_PATH_HPP
        fs::path makepath (const fs::path &p, std::size_t it)
        {
            std::stringstream s;
            s << "microstate.i" << it << "_T" << state.T << "_J" << state.J <<
                "_h" << state.h << "_EJ" << state.EJ << "_M" << state.M <<
                ".L" << L << ".i" << D << "s";
            return p / fs::path (s.str ());
        }
#endif
};

template<long long int D, long long int L> Ising<D, L>::Ising (double J,
    double h, double T, std::function<int8_t(void)> genfunc) :
        interrupt (false), running (false),
        randgen (
            std::chrono::system_clock::now ().time_since_epoch ().count ()),
        spinpicker (0, N - 1), acceptor (0, 1)
{
    MacroState st;
    std::generate (spins.begin (), spins.end (), genfunc);
    for (long long int i = 0; i < N; i++)
    {
        long long int delta = 1;
        long long int EJi = 0;
        for (long long int j = 0; j < D; j++)
        {
            EJi += spins[i] * (spins[absmod(i + delta, N)]
                    + spins[absmod(i - delta, N)]);
            delta *= L;
        }
        st.EJ += EJi / 2;
        st.M += spins[i];
    }

    st.J = J;
    st.h = h;
    st.T = T;

    state = st;

}

template<long long int D, long long int L> bool Ising<D, L>::trial (void)
{
    MacroState st = state;
    long long int i = spinpicker (randgen);
    long long int delta = 1;
    long long int EJi = 0;
    for (long long int j = 0; j < D; j++)
    {
        EJi += spins[i] * (spins[absmod (i + delta, N)]
                + spins[absmod (i - delta, N)]);
        delta *= L;
    }

    double dE = st.J * EJi + st.h*spins[i];

    if (dE < 0 || acceptor (randgen) < std::exp (-2 * dE / st.T))
    {
        st.EJ -= 2 * EJi;
        st.M -= 2 * spins[i];
        spins[i] *= -1;
        state = st;
        return true;
    }

    return false;
}

template<long long int D, long long int L> std::size_t Ising<D, L>::run (
    std::size_t iterations, std::size_t trial_limit,
    std::ostream *macrostream, std::size_t saveafter_macro,
    std::size_t saveafter, const std::string &savedir,
    std::ostream *progstream)
{
    std::size_t pct = iterations / 100;
    if (!saveafter_macro)
        saveafter_macro = 1;

    if (macrostream != nullptr)
        (*macrostream) << "0 " << state << std::endl;

#ifdef HAVE_BOOST_FILESYSTEM_PATH_HPP
    fs::path outdir (savedir);
#endif

    running = true;
    std::size_t i, last_success;
    for (i = 0; i < iterations; i++)
    {
#ifdef HAVE_BOOST_FILESYSTEM_PATH_HPP
        if (savedir.length () && (saveafter && i % saveafter == 0 || i == 0))
            save (makepath (outdir, i));
#endif
        if (trial ())
            last_success = i;

        if (progstream != nullptr && pct && i % pct == 0)
            (*progstream) << (i / pct) << "% (" << i << "/" << iterations <<
                " iterations)" << std::endl;

        if (macrostream != nullptr && i % saveafter_macro == 0)
            (*macrostream) << i + 1 << " " << state <<
                std::endl;

        if (interrupt || i - last_success == trial_limit)
            break;
    }

#ifdef HAVE_BOOST_FILESYSTEM_PATH_HPP
    if (savedir.length ())
        save (makepath (outdir, i));
#endif

    interrupt = false;
    running = false;

    return i;
}

template<long long int D, long long int L> std::size_t
Ising<D, L>::wolff (void)
{
    MacroState st = state;
    std::stack<std::size_t> trystack;
    double p = 1 - std::exp (-2 * st.J / st.T);
    std::size_t n = 0;

    auto flip = [&] (long long int i) {
        long long int delta = 1;
        for (long long int j = 0; j < D; j++)
        {
            st.EJ -= 2 * spins[i] * ((spins[absmod (i + delta, N)]
                + spins[absmod (i - delta, N)]));
            delta *= L;
        }

        spins[i] *= -1;
        trystack.push (i);
        n++;
    };

    flip (spinpicker (randgen));
    int8_t sp = -spins[trystack.top ()];

    while (trystack.size ())
    {
        long long int delta = 1;
        long long int i = trystack.top ();
        trystack.pop ();

        for (long long int j = 0; j < D; j++)
        {
            long long int k = absmod (i + delta, N);
            if (spins[k] == sp && acceptor (randgen) < p)
                flip (k);

            k = absmod (i - delta, N);
            if (spins[k] == sp && acceptor (randgen) < p)
                flip (k);

            delta *= L;
        }
    }

    st.M -= 2 * sp * n;
    state = st;

    return n;
}

template<long long int D, long long int L> std::size_t
Ising<D, L>::wolff_run (std::size_t iterations, std::ostream *macrostream,
    std::size_t saveafter, const std::string &savedir,
    std::ostream *progstream)
{
    std::size_t pct = iterations / 100;

    if (macrostream != nullptr)
        (*macrostream) << "0 " << state << std::endl;

#ifdef HAVE_BOOST_FILESYSTEM_PATH_HPP
    fs::path outdir (savedir);
#endif

    running = true;
    std::size_t i, t, ttot = 0;
    for (i = 0; i < iterations; i++)
    {
#ifdef HAVE_BOOST_FILESYSTEM_PATH_HPP
        if (savedir.length () && (saveafter && i % saveafter == 0 || i == 0))
            save (makepath (outdir, i));
#endif
        wolff ();

        if (progstream != nullptr && pct && i % pct == 0)
            (*progstream) << (i / pct) << "% (" << i << "/" << iterations <<
                " iterations)" << std::endl;

        if (macrostream != nullptr)
            (*macrostream) << i + 1 << " " << state << std::endl;

        if (interrupt)
            break;
    }

#ifdef HAVE_BOOST_FILESYSTEM_PATH_HPP
    if (savedir.length ())
        save (makepath (outdir, i));
#endif

    interrupt = false;
    running = false;

    return i;
}

template<long long int D, long long int L> void Ising<D, L>::set_J (double J)
{
    state.J = J;
}

template<long long int D, long long int L> void Ising<D, L>::set_h (double h)
{
    state.h = h;
}

template<long long int D, long long int L> void Ising<D, L>::set_T (double T)
{
    state.T = T;
}

#ifdef HAVE_BOOST_FILESYSTEM_PATH_HPP
template<long long int D, long long int L> void
Ising<D, L>::save (const fs::path &f)
{
    fs::ofstream os;
    fs::path p = f;
    std::size_t i = 0;
    while (fs::exists (p))
        p = f.parent_path () / fs::path (f.stem ().string () + "." +
            std::to_string (i++) + f.extension ().string ());
    os.open (p);

    for (auto it = spins.begin (); it != spins.end (); it++)
        os << (*it == 1 ? "+" : (*it == -1 ? "-" : "x"));
    os << std::endl;
    os.close ();
}
#endif

template<long long int D, long long int L> std::string Ising<D, L>::xyslice (
    const std::array<long long int, D == 1 ? 0 : (D - 2)> &coords,
    char upchar, char downchar)
{
    std::stringstream out;
    long long int begin = 0;
    for (long long int i = 0; i < D - 2; i++)
        begin += coords[i] * cstpow (L, i + 2);

    for (long long int y = 0; y < L; y++)
    {
        for (long long int x = 0; x < L; x++)
            out << (spins[begin + y * L + x] == 1 ? upchar :
                    (spins[begin + y * L + x] == -1 ? downchar: 'x'));

        if (y < L - 1)
            out << std::endl;
    }

    return out.str ();
}

template<long long int D, long long int L> int8_t
Ising<D, L>::inftemp_generator (void)
{
    static std::default_random_engine randgen (
        std::chrono::system_clock::now ().time_since_epoch ().count ());
    static std::uniform_int_distribution<int8_t> spindist (0, 1);

    return 2 * spindist (randgen) - 1;
}

#endif
