/* vim: se fo=tcroq : */
/* Copyright (C) 2016 Penguin Development
 *
 * This file is part of Ising, a D-dimensional Ising model simulator.
 *
 * License GPLv3+: GNU GPL version 3 or later
 *  <http://gnu.org/licenses/gpl.html>.
 * This is free software: you are free to change and redistribute it.
 * There is NO WARRANTY, to the extent permitted by law.
 *
 * Written by Peter Sterk.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifndef DIM
#define DIM 2
#endif

#ifndef GRIDSIZE
#define GRIDSIZE 32
#endif

#ifndef PACKAGE_VERSION
#define PACKAGE_VERSION "0.1"
#endif

#include <csignal>
#include <iostream>
#include <sstream>
#include <string>

#include <getopt.h>

#include <ising/ising.h>

Ising<DIM, GRIDSIZE> *ising = nullptr;

void interrupt (int sig)
{
    if (ising != nullptr)
    {
        ising->interrupt = true;
        std::cerr << "Ising::run() interrupted by signal " << sig <<
            std::endl;
    }
}

void help (const char *progname, std::ostream &s)
{
    s << "Usage:"                                           << std::endl <<
        "    " << progname << " [option...]"                << std::endl <<
                                                               std::endl <<
        "Options:"                                          << std::endl <<
        "    --temperature=T, -T T       "                  <<
        "Set the system temperature to T (in Kelvin)."      << std::endl <<
        "    --interaction=J, -J J       "                  <<
        "Set the inter-spin interaction strength to"        << std::endl <<
        "                                J*kB."             << std::endl <<
        "    --field=h, -H h             "                  <<
        "Set the external field strength to h*kB."          << std::endl <<
        "    --iterations=n, -i n        "                  <<
        "Run at most n iterations/attempted spin flips."    << std::endl <<
        "    --trial-limit=n, -l n       "                  <<
        "Accept a ground state if n successive trials"      << std::endl <<
        "                                fail."             << std::endl <<
        "    --macro-interval=n, -s n    "                  <<
        "When using the Metropolis algorithm, save"         << std::endl <<
        "                                "                  <<
        "macroscale data every n trials. Ignored when"      << std::endl <<
        "                                "                  <<
        "using the Wolff algorithm."                        << std::endl <<
        "    --save-interval=n, -s n     "                  <<
        "Save a full microstate snapshot every n"           << std::endl <<
        "                                "                  <<
        "trials. --output-dir must be set for this to"      << std::endl <<
        "                                have any effect."  << std::endl <<
        "    --output-dir=d, -o d        "                  <<
        "Save microstate data to the directory d."          << std::endl <<
        "    --load=f, -L f              "                  <<
        "Load an initial microscopic state from the"        << std::endl <<
        "                                "                  <<
        "file f. This overrides -z."                        << std::endl <<
        "    --wolff, -w                 "                  <<
        "Use the much faster Wolff algorithm. This"         << std::endl <<
        "                                "                  <<
        "algorithm demands that h=0, and is therefore"      << std::endl <<
        "                                "                  <<
        "disabled by default."                              << std::endl <<
        "    --from-zero, -z             "                  <<
        "Start the simulation at T=0 instead of"            << std::endl <<
        "                                T=infinity."       << std::endl <<
        "    --macro-stdout, -O          "                  <<
        "If -o is given, still output macrostate data"      << std::endl <<
        "                                "                  <<
        "to stdout instead of to a file."                   << std::endl <<
        "    --verbose, -v               "                  <<
        "Spit out some helpful data, including a"           << std::endl <<
        "                                "                  <<
        "progress report, to stderr."                       << std::endl <<
        "    --version, -V               "                  <<
        "Show version and exit."                            << std::endl <<
        "    --help                      "                  <<
        "Show this help message and exit."                  << std::endl;
}

void version (std::ostream &s)
{
    s << "ising" << DIM << " " << PACKAGE_VERSION           << std::endl <<
        DIM << "D component of Ising " << PACKAGE_VERSION   <<
        ": a D-dimensional Ising model simulator."          << std::endl <<
                                                               std::endl <<
        "Copyright (C) 2016 Penguin Development "           <<
        "<http://penguindevelopment.org>"                   << std::endl <<
        "License GPLv3+: GNU GPL version 3 or later "       <<
        "<http://gnu.org/licenses/gpl.html>."               << std::endl <<
        "This is free software: you are free to change "    <<
        "and redistribute it."                              << std::endl <<
        "There is NO WARRANTY, to the extent permitted "    <<
        "by law."                                           << std::endl <<
                                                               std::endl <<
        "Written by Peter Sterk."                           << std::endl;
}

int main (int argc, char **argv)
{
    struct option opts[] = {
        {"temperature",         required_argument,  NULL,   'T'},
        {"interaction",         required_argument,  NULL,   'J'},
        {"field",               required_argument,  NULL,   'H'},
        {"iterations",          required_argument,  NULL,   'i'},
        {"trial-limit",         required_argument,  NULL,   'l'},
        {"macro-interval",      required_argument,  NULL,   'm'},
        {"save-interval",       required_argument,  NULL,   's'},
        {"output-dir",          required_argument,  NULL,   'o'},
        {"load",                required_argument,  NULL,   'L'},
        {"wolff",               no_argument,        NULL,   'w'},
        {"from-zero",           no_argument,        NULL,   'z'},
        {"macro-stdout",        no_argument,        NULL,   'O'},
        {"verbose",             no_argument,        NULL,   'v'},
        {"version",             no_argument,        NULL,   'V'},
        {"help",                no_argument,        NULL,   'h'},
        {0,                     0,                  0,      0}
    };

    double T = 298.15,
           J = 1.0,
           h = 0.0;

    bool have_saveafter = false;
    std::size_t iter = ~(static_cast<std::size_t> (0));
    std::size_t trial_limit = 10000;
    std::size_t saveafter_macro = 1;
    std::size_t saveafter;

    std::string dir;
    std::string loadfile;
    std::ifstream loadstream;

    bool fromzero = false;
    bool verbose = false;
    bool wolff = false;
    bool macro_stdout = false;

    std::ostream *os = nullptr;

    for (;;)
    {
        int c = getopt_long (argc, argv, "T:J:H:i:l:a:t:m:s:o:L:wzOvVh", opts,
            NULL);
        if (c == -1)
            break;

        switch (c)
        {
            case 'T':
                T = std::stod (optarg);
                if (T <= 0)
                {
                    if (T == 0)
                        T = 0.01;
                    else
                        T *=-1;
                    std::cerr << "T must be greater than 0; setting T="
                        << T << std::endl;
                }
                break;
            case 'J':
                J = std::stod (optarg);
                break;
            case 'H':
                h = std::stod (optarg);
                break;
            case 'i':
                iter = std::stoull (optarg);
                if (iter == 0)
                    iter = ~(static_cast<std::size_t> (0)) - 1;
                break;
            case 'l':
                trial_limit = std::stoull (optarg);
                if (trial_limit == 0)
                    trial_limit = 1;
                break;
            case 'm':
                saveafter_macro = std::stoull (optarg);
                break;
            case 's':
                saveafter = std::stoull (optarg);
                have_saveafter = true;
                break;
            case 'o':
#ifdef HAVE_BOOST_FILESYSTEM_PATH_HPP
                dir = std::string (optarg);
#else
                std::cerr << "Error: --output-dir not supported, as Ising " <<
                    "was compiled without" << std::endl <<
                    "       boost::filesystem support" << std::endl;
                return 1;
#endif
                break;
            case 'L':
                loadfile = std::string (optarg);
                break;
            case 'w':
                wolff = true;
                break;
            case 'z':
                fromzero = true;
                break;
            case 'O':
                macro_stdout = true;
            case 'v':
                verbose = true;
                break;
            case 'V':
                version (std::cout);
                return 0;
                break;
            case 'h':
                help (argv[0], std::cout);
                return 0;
                break;
            default:
                help (argv[0], std::cerr);
                return 1;
                break;
        }
    }

    if (!have_saveafter)
        saveafter = iter / 10;

    auto genstart = std::chrono::system_clock::now ();
    std::function<int8_t(void)> genfunc;
    if (loadfile.length ())
    {
        loadstream.open (loadfile);
        genfunc = Ising<DIM, GRIDSIZE>::StreamLoaderGenerator (loadstream);
    }
    else
    {
        genfunc = fromzero ?
                Ising<DIM, GRIDSIZE>::zerotemp_generator :
                Ising<DIM, GRIDSIZE>::inftemp_generator;
    }
    ising = new Ising<DIM, GRIDSIZE> (J, h, T, genfunc);
    auto genend = std::chrono::system_clock::now ();

#ifdef HAVE_BOOST_FILESYSTEM_PATH_HPP
    fs::path macropath;
    if (dir.length () && !macro_stdout)
    {
        fs::path p (dir);
        if (fs::exists (p))
        {
            if (!fs::is_directory (p))
            {
                std::cerr << "Error: " << p << " exists, but is not a " <<
                    "directory." << std::endl;
                return 1;
            }
        }
        else
        {
            fs::create_directories (p);
        }

        auto state = ising->get_macrostate ();
        std::stringstream s;
        s << "macrostate.T" << state.T << "_J" << state.J << "_h" <<
            state.h << ".L" << ising->size << ".i" << ising->dim << "M";

        fs::path mpname = fs::path (s.str ());
        macropath = p / mpname;
        std::size_t i = 0;
        while (fs::exists (macropath))
            macropath = p / fs::path (mpname.stem ().string () + "." +
                std::to_string (i++) + mpname.extension ().string ());

        os = new fs::ofstream (macropath);
    }
#endif

    if (os == nullptr)
        os = &std::cout;

    if (verbose)
    {
        std::cerr << "Generated " << ising->N << " spins in " <<
            (std::chrono::duration_cast<std::chrono::microseconds> (genend -
                genstart)).count () << " µs." << std::endl <<
            "Starting " << (wolff ? "Wolff" : "Metropolis") <<
            " algorithm simulation of max " << iter << " iterations." <<
            std::endl << "Initial state " << ising->get_macrostate () <<
            std::endl;
        if (wolff)
            std::cerr << "Simulation will halt if " << trial_limit <<
            " consecutive trials are unsuccessful." << std::endl;
        std::cerr << "Macrostate data will be output to ";
#ifdef HAVE_BOOST_FILESYSTEM_PATH_HPP
        if (dir.length () && !macro_stdout)
            std::cerr << macropath << "." << std::endl;
        else
#endif
            std::cerr << "stdout." << std::endl;
    }

    std::signal (SIGINT, interrupt);
#ifdef SIGHUP
    std::signal (SIGHUP, interrupt);
#endif

    auto simstart = std::chrono::system_clock::now ();
    std::size_t iters;
    if (wolff)
    {
        if (h != 0)
        {
            std::cerr << "The Wolff algorithm can only be used in " <<
                "field-free (h=0) mode, but h=" << h << " was requested." <<
                std::endl;
            return 1;
        }
        iters = ising->wolff_run (iter, os, saveafter, dir, verbose ?
            &std::cerr : nullptr);
    }
    else
    {
        iters = ising->run (iter, trial_limit, os, saveafter_macro, saveafter,
            dir, verbose ? &std::cerr : nullptr);
    }
    auto simend = std::chrono::system_clock::now ();
    std::signal (SIGINT, SIG_DFL);
#ifdef SIGHUP
    std::signal (SIGHUP, SIG_DFL);
#endif

    if (verbose)
        std::cerr << "Completed " << iters << " iterations in " <<
            (std::chrono::duration_cast<std::chrono::seconds> (
                simend - simstart)).count () << " s (total " <<
            (std::chrono::duration_cast<std::chrono::seconds> (
                simend - genstart)).count () << "s)." << std::endl <<
            "Final state: " << ising->get_macrostate () << std::endl;

#ifdef HAVE_BOOST_FILESYSTEM_PATH_HPP
    if (dir.length ())
    {
        (dynamic_cast<fs::ofstream *> (os))->close ();
        delete os;
    }
#endif

    delete ising;

    return 0;
}
